/**********************************************************************
 * handle.c                                                October 2002
 * Horms                                             horms@verge.net.au
 *
 * Encapsulation of an open connection to a Multicast Socket
 *
 * libvaness_mcast
 * Multicast Helper Library
 * Copyright (C) 2000 Horms
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 **********************************************************************/


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <vanessa_logger.h>
#include <vanessa_socket.h>

#include "vanessa_mcast.h"
#include "vanessa_mcast.h"


/**********************************************************************
 * vanessa_mcast_create
 * Create a vanessa_mcast_t
 *
 * pre: none
 * post: new vanessa_mcast_t is allocated and initialised
 * return: pointer to new vanessa_mcast_t
 **********************************************************************/

vanessa_mcast_t *vanessa_mcast_create(void)
{
	vanessa_mcast_t *hdl;

	hdl = (vanessa_mcast_t *)calloc(1, sizeof(vanessa_mcast_t));
	if(!hdl) {
		VANESSA_LOGGER_DEBUG_ERRNO("calloc");
		return(NULL);
	}

	return(hdl);
}


/**********************************************************************
 * vanessa_mcast_destroy
 * Destroy a vanessa_mcast_t
 *
 * pre: hdl: vanessa_mcast_t to destroy
 * post: hdl is freed
 * return: none
 **********************************************************************/

void vanessa_mcast_destroy(vanessa_mcast_t *hdl) 
{
	free(hdl);
}


/**********************************************************************
 * vanessa_mcast_open
 * Open a socket for a vanessa_mcast_t
 * Does not bind connect or set any socket options
 * Intended to be called by vanessa_mcast_open_server()
 * and vanessa_mcast_open_client()
 *
 * pre: group: group to open
 *      port: port to use
 *      flags: flags passed to vanessa_socket_host_port_sockaddr_in()
 * post: New vanessa_mcast_t is created
 *       socket is opened for group and port and assigned to the new handle
 * return: new handle
 *         NULL on error
 **********************************************************************/

static vanessa_mcast_t *vanessa_mcast_open(const char *group, const char *port, 
		unsigned int flags)
{
	vanessa_mcast_t *hdl;

	if(vanessa_mcast_group_test(group, flags) < 1) {
		VANESSA_LOGGER_DEBUG("not a mulicast group");
		return(NULL);
	}

	hdl = vanessa_mcast_create();
	if(!hdl) {
		VANESSA_LOGGER_DEBUG("vanessa_mcast_create");
		return(NULL);
	}

	if(vanessa_socket_host_port_sockaddr_in(group, port,
			&(VANESSA_MCAST_GET_SA_IN(hdl)), flags) < 0) {
		VANESSA_LOGGER_DEBUG("vanessa_socket_host_port_sockaddr_in");
		return(NULL);
	}

	/* Create socket */
	VANESSA_MCAST_GET_SOCKET(hdl) = socket(AF_INET, SOCK_DGRAM, 0);
	if(VANESSA_MCAST_GET_SOCKET(hdl) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("socket");
		return(NULL);
	}
	
	return(hdl);
}


/**********************************************************************
 * vanessa_mcast_get_ifaddr
 * Get an interface address specified by the ifname argument.
 *
 * pre: hdl: handle used to get the interface address.
 *      ifname: interface name.
 * post: addr: interface address.
 * return: 0 on success
 *         <0 on error
 **********************************************************************/

static int vanessa_mcast_get_ifaddr(vanessa_mcast_t *hdl,
				    const char *ifname, struct in_addr *addr)
{
	int s;
	int err;
	struct ifreq ifreq;

	s = VANESSA_MCAST_GET_SOCKET(hdl);
	memset(&ifreq, 0, sizeof(ifreq));
	strncpy(ifreq.ifr_name, ifname, IFNAMSIZ);

	if((err = ioctl(s, SIOCGIFADDR, &ifreq)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("ioctl");
	} else {
		*addr = ((struct sockaddr_in *)&ifreq.ifr_addr)->sin_addr;
	}
	return(err);
}


/**********************************************************************
 * vanessa_mcast_open_server
 * vanessa_mcast_t for listening as a server
 *
 * pre: group: group to open
 *      port: port to use
 *      ifname: interface to listen on
 *      flags: flags passed to vanessa_mcast_open()
 * post: Handle is created
 *       Socket is opened
 *       Socket is set as reusable
 *       Socket is bound to group ad port
 *       Socket joins Group
 *          If ifname is set, then it joins the group on that interface
 *          Else on the default interface as per the kernel
 * return: new handle
 *         NULL on error
 **********************************************************************/

vanessa_mcast_t *vanessa_mcast_open_server(const char *group, const char *port, 
		const char *ifname, unsigned int flags) 
{
	int one;
	int s;
	struct ip_mreq mreq;
	vanessa_mcast_t *hdl;
	struct sockaddr_in sa;

	/* Open socket */
	hdl = vanessa_mcast_open(group, port, flags);
	if(!hdl) {
		VANESSA_LOGGER_DEBUG("open_multicast_socket");
		return(NULL);
	}

	s = VANESSA_MCAST_GET_SOCKET(hdl);
	sa = VANESSA_MCAST_GET_SA_IN(hdl);

	/* Allow port to be reused */
	one = 1;
	if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *)&one, 
				sizeof(one)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("setsockopt SO_REUSEADDR");
		close(s);
		return(NULL);
	}

	/* Bind socket */
	if(bind(s, (struct sockaddr *) &sa, sizeof(sa)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("bind");
		close(s);
		return(NULL);
	}

	/* Join Group */
	memset(&mreq, 0, sizeof(mreq));

	mreq.imr_multiaddr.s_addr=sa.sin_addr.s_addr;

	if(ifname) {
		if(vanessa_mcast_get_ifaddr(hdl, ifname,
				&mreq.imr_interface) < 0) {
			close(s);
			return(NULL);
		}
	}
	else {
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	}

	if(setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq,
			sizeof(mreq)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("setsockopt IP_ADD_MEMBERSHIP");
		close(s);
		return(NULL);
	}

	return(hdl);
}


/**********************************************************************
 * vanessa_mcast_open_client
 * vanessa_mcast_t for requesting as a client
 *
 * pre: group: group to open
 *      port: port to use
 *      ifname: interface to listen on
 *      ttl: Time to live for outoging dagrams (in hops)
 *      loopback: Receive packets from sent by this host
 *                0 for no, 1 for yes
 *      flags: flags passed to vanessa_mcast_open()
 * post: Handle is created
 *       Socket is opened
 *       Set loopback option for socket
 *       Set ttl for socket
 * return: new handle
 *         NULL on error
 **********************************************************************/

vanessa_mcast_t *vanessa_mcast_open_client(const char *group, const char *port, 
		const char *ifname, unsigned char ttl, unsigned int loopback, 
		unsigned int flags)
{
	int s;
	int tmp;
	vanessa_mcast_t *hdl;

	/* Open socket */
	hdl = vanessa_mcast_open(group, port, flags);
	if(!hdl) {
		VANESSA_LOGGER_DEBUG("open_multicast_socket");
		return(NULL);
	}

	s = VANESSA_MCAST_GET_SOCKET(hdl);

	/* Turn loopback off */
	tmp = loopback;
	if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_LOOP, 
				(void *) &tmp, sizeof(tmp)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("setsockopt IP_MULTICAST_LOOP");
		close(s);
		return(NULL);
	}

	/* Set TTL */
	tmp = ttl;
	if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL, (void *)&tmp, 
				sizeof(tmp)) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("setsockopt IP_MULTICAST_TTL");
		close(VANESSA_MCAST_GET_SOCKET(hdl));
		return(NULL);
	}

	if(ifname && vanessa_mcast_set_iface(hdl, ifname) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("vanessa_mcast_set_iface");
		close(VANESSA_MCAST_GET_SOCKET(hdl));
		return(NULL);
	}

	return(hdl);
}


/**********************************************************************
 * vanessa_mcast_close
 * Close a handle by closing its file descriptor and freeing its memory
 *
 * pre: hdl: handle to close
 * return: 0 on success
 *         < 0 on error, as per close(2)
 **********************************************************************/

int vanessa_mcast_close(vanessa_mcast_t *hdl) 
{
	int status;

	status = close(VANESSA_MCAST_GET_SOCKET(hdl));
	vanessa_mcast_destroy(hdl);

	return(status);
}



/**********************************************************************
 * __vanessa_mcast_sendto
 * Send data to a handle. The guts of vanessa_mcast_send()
 * and vanessa_mcast_sendto()
 *
 * pre: hdl: handle to send data to
 *      msg: message to send
 *      len: length of message in bytes
 *      flags: passed to send(2) or sendto(2)
 *      internal_flags: should be __VANESSA_MCAST_SENDTO or 
 *                      __VANESSA_MCAST_SEND which will result in 
 *                      sendto(2) or send() being called respectively.
 * post: send(2) or sendto(2) will be called repeatedly until either
 *       len bytes of msg is sent, or an error occurs
 * return: len on success
 *         < 0, as per sendto(2) or send(2)
 **********************************************************************/

#define __VANESSA_MCAST_SENDTO 1
#define __VANESSA_MCAST_SEND   2

static int __vanessa_mcast_sendto(vanessa_mcast_t *hdl, const void *msg, 
		size_t len, int flags, int internal_flags)
{
	int sk;
	int status;
	int do_send;
	size_t offset;
	struct sockaddr *sa = NULL;
	socklen_t salen = 0;

	do_send = (internal_flags == __VANESSA_MCAST_SEND) ? 1 : 0;

	offset = 0;
	sk = VANESSA_MCAST_GET_SOCKET(hdl);
	if(internal_flags != __VANESSA_MCAST_SEND) {
		sa = (struct sockaddr *)&VANESSA_MCAST_GET_SA_IN(hdl);
		salen = sizeof(*sa);
	}

        /* Write as much as we can until there
	 * is nothing left, we can't write any more or
	 * there is an error */
	while(1) {
		if(do_send) {
			status = send(sk, msg+offset, len-offset, flags);
		}
		else {
			status = sendto(sk, msg+offset, len-offset, flags, 
					sa, salen);
		}
		offset += status;
		if(status < 0) {
			VANESSA_LOGGER_DEBUG_ERRNO(do_send?"send":"sendto");
			return(status);
		}
		if(!status) {
			return(0);
		}
		if(offset >= len) {
			break;
		}
	}

	return(len);
}


/**********************************************************************
 * vanessa_mcast_sendto
 * Send data to a handle, analoglous to sendto(2)
 *
 * pre: hdl: handle to send data to
 *      msg: message to send
 *      len: length of message in bytes
 *      flags: passed to sendto(2)
 * post: sendto(2) will be called repeatedly until either
 *       len bytes of msg is sent, or an error occurs
 * return: len on success
 *         < 0, as per sendto(2)
 **********************************************************************/

int vanessa_mcast_sendto(vanessa_mcast_t *hdl, const void *msg, size_t len, 
		int flags)
{
	int status;

	status = __vanessa_mcast_sendto(hdl, msg, len, flags,
			__VANESSA_MCAST_SENDTO);
	if (status < 0) {
		VANESSA_LOGGER_DEBUG("__vanessa_mcast_sendto");
		return(status);
	}

	return(status);
}


/**********************************************************************
 * vanessa_mcast_send
 * Send data to a handle, analoglous to send(2)
 *
 * pre: hdl: handle to send data to
 *      msg: message to send
 *      len: length of message in bytes
 *      flags: passed to send(2)
 * post: send(2) will be called repeatedly until either
 *       len bytes of msg is sent, or an error occurs
 * return: len on success
 *         < 0, as per send(2)
 **********************************************************************/

int vanessa_mcast_send(vanessa_mcast_t *hdl, const void *msg, size_t len, 
		int flags)
{
	int status;

	status = __vanessa_mcast_sendto(hdl, msg, len, flags,
			__VANESSA_MCAST_SEND);
	if (status < 0) {
		VANESSA_LOGGER_DEBUG("__vanessa_mcast_sendto");
		return(status);
	}

	return(status);
}


/**********************************************************************
 * vanessa_mcast_recvfrom
 * Receive data from a handle, analoglous to recvfrom(2)
 *
 * pre: hdl: handle to receive data from
 *      msg: buffer to receive message into
 *      len: length of message in bytes
 *      flags: passed to recvfrom(2)
 *      from: handle to recieve from
 *            may be NULL, in which case behaviour is the
 *            same as vanessa_mcast_recvfrom()
 * post: recvfrom(2) is called once to read data
 * return: number of bytes read
 *         < 0, as per recvfrom(2)
 **********************************************************************/

int vanessa_mcast_recvfrom(vanessa_mcast_t *hdl, void *msg, size_t len, 
		int flags, vanessa_mcast_t *from)
{
	int status;
	struct sockaddr_in *sa;
	socklen_t salen;

	salen = sizeof(*sa);
	sa = (from) ? &VANESSA_MCAST_GET_SA_IN(from) : NULL;

	status = recvfrom(VANESSA_MCAST_GET_SOCKET(hdl), msg, len, flags,
			(struct sockaddr *)sa, &salen);
	if(status < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("recvfrom");
		return(status);
	}

	return(status);
}


/**********************************************************************
 * vanessa_mcast_recvfrom
 * Receive data from a handle, analoglous to recv(2)
 *
 * pre: hdl: handle to receive data from
 *      msg: buffer to receive message into
 *      len: length of message in bytes
 *      flags: passed to recvfrom(2)
 * post: recvfrom(2) is called once to read data
 * return: number of bytes read
 *         < 0, as per recvfrom(2)
 **********************************************************************/

int vanessa_mcast_recv(vanessa_mcast_t *hdl, void *msg, size_t len, int flags)
{
	int status;

	status = vanessa_mcast_recvfrom(hdl, msg, len, 0, NULL);
	if(status < 0) {
		VANESSA_LOGGER_DEBUG("vanessa_mcast_recvfrom");
		return(status);
	}

	return(status);
}

/**********************************************************************
 * vanessa_mcast_set_iface
 * set sending interface of vanessa_mcast.
 *
 * pre: hdl: handle to set sending interface
 *      ifname: interface name to be set to the handle
 * return: 0 on success
 *         <0 on error
 **********************************************************************/

int vanessa_mcast_set_iface(vanessa_mcast_t *hdl, const char *ifname)
{
	int err;
	struct ip_mreqn mreq;

	memset(&mreq, 0, sizeof(mreq));
	if((err = vanessa_mcast_get_ifaddr(hdl, ifname,
				&mreq.imr_address)) < 0) {
		return err;
	}

	if((err = setsockopt(VANESSA_MCAST_GET_SOCKET(hdl),
				IPPROTO_IP, IP_MULTICAST_IF,
				(void *)&mreq, sizeof(mreq))) < 0) {
		VANESSA_LOGGER_DEBUG_ERRNO("setsockopt IP_MULTICAST_IF");
	}

	return err;
}

